from ._printing import colorize, spinner, format_table, goto_xy, progress_bar, print_table  # noqa: F401, F403, E501
from ._keyhandler import KeyHandler, Keys, getch  # noqa: F401

__all__ = [
    'colorize',
    'spinner',
    'format_table',
    'print_table',
    'goto_xy',
    'progress_bar',
    'KeyHandler',
    'Keys',
    'getch'
]



import argparse
import asyncio
import functools
import itertools
import logging
import os
import sys
import typing

import colorama
import stringcase
import pkg_resources

from ._services import Service
from .events import EventDispatcher, PreceptEvent
from ._configs import Config, config_factory
from ._tools import is_windows
from ._cli import CombinedFormatter, Cli, Argument, Command
from ._executor import AsyncExecutor
from ._logger import setup_logger

from ._cli import CommandMeta


class PreceptMeta(CommandMeta):
    def __new__(mcs, name, bases, attributes):
        new_attributes = dict(**attributes)
        prog_name = attributes.get('_prog_name')
        new_attributes['_prog_name'] = prog_name or stringcase.spinalcase(name)
        # pylint: disable=too-many-function-args
        return CommandMeta.__new__(mcs, name, bases, new_attributes)


class Precept(metaclass=PreceptMeta):
    """
    Auto cli generator, methods decorated with ``Command`` will have
    a corresponding sub-command in the cli application.

    Commands will get the arguments named as the last element of the command
    flags.

    Override `main` method for root handler, it gets all the `global_arguments`
    """
    _commands = []
    prog_name = ''
    global_arguments = []
    default_configs: dict = {}
    version = '0.0.1'
    config_class = None
    config: Config = None

    # pylint: disable=too-many-locals
    def __init__(
            self,
            config_file: typing.Union[str, typing.List[str]] = None,
            loop=None,
            executor=None,
            executor_max_workers=None,
            add_dump_config_command=False,
            help_formatter=CombinedFormatter,
            logger_level=logging.INFO,
            logger_fmt=None,
            logger_datefmt=None,
            logger_stream=sys.stderr,
            logger_colors=None,
            logger_style='%',
            services: typing.List[Service] = None,
            print_version: bool = True,
    ):
        """
        :param config_file: Path to the default config file to use. Can be
            specified with ``--config-file``
        :param loop: Asyncio loop to use.
        :param executor: concurrent executor to use.
        :param add_dump_config_command: Add a ``dump-config`` command.
        :param help_formatter: The cli formatter to use.
        :param logger_level: Set logger level when setting up logging.
        :param logger_fmt: The format of the logger.
        :param logger_datefmt: Date format of the logger.
        :param logger_stream: The stream to print the logs.
        :param logger_colors: Dictionary with key logger level name and values
            of bg/fg/style dict.
        :param logger_style: The symbol to use for formatting.
        :param services: List of global services to start with the program.
        :param print_version: Print the version & name of the app before start.
        """
        self.prog_name = self.prog_name or stringcase.spinalcase(
            self.__class__.__name__
        )
        self._config_file = config_file
        if not isinstance(self._config_file, list)\
                and isinstance(config_file, str):
            self._config_file = [config_file]
        self._user_configs = None
        self.services = services or []
        self._command = None
        self._args = {}
        self.print_version = print_version

        if is_windows():  # pragma: no cover
            colorama.init()

        self.logger = setup_logger(
            self.prog_name,
            logger_level,
            logger_fmt,
            logger_datefmt,
            logger_stream,
            logger_colors,
            style=logger_style
        )
        self.executor = AsyncExecutor(
            loop, executor, max_workers=executor_max_workers
        )
        self.loop = self.executor.loop
        self.events = EventDispatcher()
        self.plugins = {}

        common_g_arguments = [
            Argument('-v', '--verbose', action='store_true', default=False),
            Argument('--log-file', type=argparse.FileType('w')),
            Argument('--quiet', action='store_true'),
        ]

        if config_file:
            common_g_arguments.append(
                Argument(
                    '-c', '--config-file',
                    type=str,
                    help='Config file path'
                )
            )

        if self.config is None:
            if self.config_class and self.config:
                # pylint: disable=not-callable
                self.config: Config = self.config_class()
            elif self.default_configs:
                cls = config_factory(self.default_configs)
                self.config = cls()
            else:
                self.config = Config()

        # Insert global arguments from config
        for _, _, prop in self.config.get_prop_paths():
            if prop.auto_global:
                key = f'--{stringcase.spinalcase(prop.qualified_name)}'
                common_g_arguments.append(
                    Argument(
                        key,
                        type=prop.config_type,
                        default=prop.default,
                        help=prop.comment,
                    )
                )

        # Gather commands
        attributes = dir(self)
        commands = list(
            itertools.chain(*(
                # Don't go into descriptors yet, class members gets the
                getattr(self.__class__, x).get_commands()
                for x in self._commands if x in attributes
            ))
        )

        if add_dump_config_command:
            @Command(
                Argument(
                    'outfile',
                    help='Write the current configs to this file.',
                    type=str,
                    default=config_file,
                ),
                description='Dump the current configuration file content.'
            )
            async def dump_configs(outfile):
                dirname = os.path.dirname(outfile)
                if dirname:
                    os.makedirs(dirname, exist_ok=True)
                self.config.save(outfile)

            commands.append((
                dump_configs.command.command_name,
                dump_configs.command,
                dump_configs
            ))

        self.cli = Cli(
            *[
                (
                    x[0],
                    x[1],
                    # Now go into the descriptors for that self argument.
                    getattr(self, x[1].obj_name)
                    if x[1].obj_name in attributes else x[2]
                ) for x in commands
            ],
            prog=self.prog_name,
            description=getattr(self, '__doc__', ''),
            global_arguments=common_g_arguments + self.global_arguments,
            on_parse=self._on_parse,
            default_command=self.main,
            formatter_class=help_formatter,
            events=self.events
        )

        setattr(self.config, '_app', self)

        if not self.loop.is_running():
            self.loop.run_until_complete(self.setup_plugins())
        else:
            self.logger.warning(
                'Precept application instantiated '
                'from a running loop. No plugins auto setup.'
            )

    @property
    def config_path(self):
        if self._user_configs:
            return self._user_configs
        for config in self._config_file:
            if os.path.exists(config):
                return config
        return ''

    def start(self, args=None):
        """
        Start the application loop.

        :return:
        """
        self.loop.run_until_complete(
            self.events.dispatch(str(PreceptEvent.BEFORE_CLI_START))
        )
        self.loop.run_until_complete(self.cli.run(args=args))
        self.loop.run_until_complete(
            self.stop_services(
                command=getattr(
                    getattr(self.__class__, self._command, None),
                    'command',
                    None
                )
            )
        )
        self.loop.run_until_complete(
            self.events.dispatch(str(PreceptEvent.CLI_STOPPED))
        )

    async def setup_services(self, command: Command = None):
        """
        Setup the services for the command or the main application.

        :param command: The command that was run.
        :return:
        """
        for service in itertools.chain(
                self.services, (command and command.services or [])
        ):
            await service.setup(self)

    async def start_services(self, command: Command = None):
        """
        Start the services, automatically called by start.

        If the application if run with another method you can call this to
        start the global services without the command argument.

        :param command: The command that was run.
        :return:
        """
        services = []
        for service in itertools.chain(
                self.services, (command and command.services or [])
        ):
            if not service.running:
                task = self.loop.create_task(service.start())
                task.add_done_callback(
                    functools.partial(
                        lambda x: self.logger.debug(
                            f'Started service {x}'
                        ),
                        service.name
                    )
                )
                services.append(task)

        await asyncio.gather(*services)

    async def stop_services(self, command: Command = None):
        """
        Stop the services, automatically called by start.

        Call this if your application is not run with start
        and you have running services.

        :param command: The command that was run.
        :return:
        """
        services = []
        for service in itertools.chain(
                self.services, (command and command.services or [])
        ):
            if service.running:
                task = self.loop.create_task(service.stop())
                task.add_done_callback(
                    functools.partial(
                        lambda x: self.logger.debug(
                            f'Stopped service {x}'
                        ),
                        service.name
                    )
                )
                services.append(task)

        await asyncio.gather(*services)

    async def setup_plugins(self):
        """
        Load and setup the registered plugins.

        To register a plugin, subclass ``Plugin`` and instantiate
        then add to ``setup.py`` entry_points:

            '{app_name}.plugins': ['my_plugin = plugin_module:plugin']

        :return:
        """
        for plugin in pkg_resources.iter_entry_points(
                f'{stringcase.snakecase(self.prog_name)}.plugins'
        ):
            plug = plugin.load()
            await plug.setup(self)
            self.plugins[plugin.name] = plug

    # pylint: disable=unused-argument
    async def main(self, **kwargs):
        """
        Handler when no command has been entered. Gets the globals arguments.

        :param kwargs: Global arguments.
        :return:
        """
        self.logger.error('Please enter a command')
        self.cli.parser.print_help()

    async def _on_parse(self, args):
        if args.verbose:
            self.logger.setLevel(logging.DEBUG)

        if args.log_file:
            self.logger.addHandler(logging.StreamHandler(args.log_file))

        if args.quiet:
            self.logger.setLevel(logging.ERROR)

        if self._config_file:
            if args.config_file:
                self._user_configs = args.config_file

            if self.config_path:
                self.logger.info(f'Using config {self.config_path}')
                self.config.read_file(self.config_path)

        await self.events.dispatch(
            str(PreceptEvent.CLI_PARSED),
            arguments=args
        )

        self._args = args
        self._command = args.command or ''

        command = getattr(
            getattr(self.__class__, self._command, None),
            'command',
            None
        )
        await self.setup_services(command)
        await self.start_services(command)

        if self.print_version:
            self.logger.info(f'{self.prog_name} {self.version}')



import collections
import abc
import itertools
import json
import os
import typing
import configparser
import textwrap
from enum import auto

import stringcase
import tomlkit

from ruamel import yaml
from ruamel.yaml.comments import CommentedMap

from ._tools import AutoNameEnum
from .errors import ConfigError


undefined = object()


def get_deep(data, *keys, default=None):
    value = data
    found = False
    keys = iter(keys)

    while not found:
        try:
            key = next(keys)
            value = value.get(key, undefined)
            if value is undefined:  # pragma: no cover
                return default, False
        except StopIteration:
            return value, True

    return value, found  # pragma: no cover


def _to_yaml(root: CommentedMap, obj):
    if isinstance(obj, Nestable):
        data = CommentedMap()
        for i, prop in enumerate(
                getattr(type(obj), x) for x in getattr(obj, '_props', [])
        ):
            root.insert(
                i, prop.name, _to_yaml(data, getattr(obj, prop.name)),
                comment=prop.comment
            )
        return root
    return obj


def _to_dict(obj):
    if isinstance(obj, Nestable):
        data = {}
        for k, v in obj.items():
            data[k] = _to_dict(v)
        return data
    return obj


class BaseConfigSerializer:

    def dump(self, configs, path):  # pragma: no cover
        raise NotImplementedError

    def load(self, path):  # pragma: no cover
        raise NotImplementedError


class YamlConfigSerializer(BaseConfigSerializer):
    def dump(self, configs, path):
        ya_data: CommentedMap = CommentedMap()
        root_comment = getattr(configs, '__doc__', None)
        if root_comment:
            ya_data.yaml_set_start_comment(root_comment)

        ya_data = _to_yaml(ya_data, configs)

        with open(path, 'w') as f:
            yaml.dump(ya_data, f, Dumper=yaml.RoundTripDumper)

    def load(self, path):
        with open(path, 'r') as f:
            return yaml.load(f, Loader=yaml.RoundTripLoader)


class JsonConfigSerializer(BaseConfigSerializer):
    def dump(self, configs, path):
        with open(path, 'w') as f:
            json.dump(_to_dict(configs), f)

    def load(self, path):  # pragma: no cover
        with open(path, 'r') as f:
            return json.load(f)


class IniConfigSerializer(BaseConfigSerializer):
    def __init__(self, root_name):
        self.root_name = root_name

    def dump(self, configs, path):
        cfg = configparser.ConfigParser(allow_no_value=True)
        leftovers = []

        root_comment = getattr(configs, '__doc__', '')
        if root_comment:
            cfg.setdefault(self.root_name, {})
            for comment in root_comment.split(os.linesep):
                cfg.set(self.root_name, f'# {comment}', None)

        for p, value, prop in configs.get_prop_paths():
            if '.' in p:
                top = '.'.join(p.split('.')[:-1])
            else:
                top = self.root_name

            cfg.setdefault(top, {})

            if isinstance(value, Nestable):
                # Put it before the first value
                if prop.comment:
                    leftovers = prop.comment.split(os.linesep)
                continue

            if prop.comment:
                for c in itertools.chain(*[
                        leftovers, prop.comment.split(os.linesep)
                ]):
                    cfg.set(top, f'# {c}', None)
                leftovers = []

            if isinstance(value, list):
                cfg[top][prop.name] = yaml.round_trip_dump(value)
            else:
                cfg[top][prop.name] = str(value)

        with open(path, 'w') as f:
            cfg.write(f)

    def load(self, path):
        cfg = configparser.ConfigParser()
        with open(path) as f:
            cfg.read_file(f)

        # noinspection PyProtectedMember
        raw = dict(cfg._sections)
        data = raw.pop(self.root_name)
        for k, v in raw.items():
            # The rest are Nestables.
            sections = k.split('.')
            key = sections[-1]
            last_section = data
            for section in sections[:-1]:
                last_section = last_section.setdefault(section, {})

            last_section[key] = v

        return data


class TomlConfigSerializer(BaseConfigSerializer):

    def dump(self, configs, path):
        doc = tomlkit.document()
        root_comment = getattr(configs, '__doc__', '')

        def add_comment(sec, comment):
            for line in textwrap.wrap(comment.strip()):
                sec.add(tomlkit.comment(line))

        if root_comment:
            add_comment(doc, root_comment.strip())
            doc.add(tomlkit.nl())

        for p, value, prop in configs.get_prop_paths():
            section = doc
            key = p
            if '.' in p:
                parts = p.split('.')
                key = parts[-1]

                for part in parts[:-1]:
                    section = section[part]

            if isinstance(value, Nestable):
                # Just add a table for those.
                table = tomlkit.table()
                section.add(key, table)
                if prop.comment is not None:
                    add_comment(table, prop.comment)
                    table.add(tomlkit.nl())
            else:
                if prop.comment is not None:
                    if len(prop.comment) > 50:
                        section.add(tomlkit.nl())
                        section.add(key, value)
                        add_comment(section, prop.comment)
                    else:
                        section.add(key, value)
                        section[key].comment(prop.comment)
                else:
                    section.add(key, value)

        with open(path, 'w') as file:
            file.write(tomlkit.dumps(doc))

    def load(self, path):
        with open(path) as file:
            return tomlkit.parse(file.read())


class ConfigFormat(AutoNameEnum):
    """
    Available formats to use with configs.

    - TOML provided by tomlkit, supports comments and more complex types.
    - YML provided by ruamel.yaml, supports comments and more complex types.
    - JSON stdlib, no support for comments and types.
    - INI stdlib, support for comments.
    """
    TOML = auto()
    YML = auto()
    JSON = auto()
    INI = auto()

    def serializer(self, config):
        if self == ConfigFormat.TOML:
            return TomlConfigSerializer()
        if self == ConfigFormat.YML:
            return YamlConfigSerializer()
        if self == ConfigFormat.JSON:
            return JsonConfigSerializer()
        if self == ConfigFormat.INI:
            return IniConfigSerializer(config.root_name)
        raise ConfigError('Invalid config format')  # pragma: no cover


class ConfigProperty:
    def __init__(
            self,
            default=None,
            comment=None,
            config_type=None,
            environ_name=None,
            auto_environ=True,
            name=None,
            auto_global=False,
    ):
        self.default = default
        self.comment = comment
        self.name = name
        self.qualified_name = None
        self.config_type = config_type or type(default) if default else None
        self.environ_name = environ_name
        self.auto_environ = auto_environ
        self.auto_global = auto_global

    def __set_name__(self, owner, name):
        self.name = name
        if not issubclass(owner, Config):
            self.qualified_name = f'{owner.__name__.lower()}.{name}'
        else:
            self.qualified_name = name

        if self.environ_name is None and self.auto_environ:
            self.environ_name = name.upper()

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = undefined

        app = getattr(instance, '_app', None)

        if self.auto_global and app is not None:
            value = app.cli.globals.get(self.qualified_name, undefined)

        if self.environ_name and value is undefined:
            # pylint: disable=invalid-envvar-default
            value = os.getenv(self.environ_name, undefined)

        if value is undefined:
            if issubclass(owner, Config):
                value = instance.get(self.name, self.default)
            else:
                root, levels = instance.get_root(self.name)
                value, found = get_deep(root, *levels)
                if not found:  # pragma: no cover
                    value = self.default

        if value is not None and self.config_type is not None:
            try:
                if isinstance(value, str) and self.config_type == list:
                    value = yaml.round_trip_load(value)
                else:
                    value = self.config_type(value)
            except TypeError as err:
                raise ConfigError(
                    f'Expected type {self.config_type.__name__} for {value}'
                ) from err

        return value

    def __set__(self, instance, value):
        if isinstance(instance, Config):
            # noinspection PyProtectedMember
            instance._data[self.name] = value
        else:
            root, levels = instance.get_root(self.name)
            current = root
            # Don't take the last level as it's the name of the value
            # we want to set.
            for level in list(levels)[:-1]:
                current = current[level]
            current[self.name] = value

    def __repr__(self):
        return f'<ConfigProperty {self.name}>'


class ConfigMeta(abc.ABCMeta):
    # pylint: disable=arguments-differ
    def __new__(mcs, name, bases, attributes):
        _new = attributes.copy()
        _props = list(itertools.chain(*(
            getattr(b, '_props', []) for b in bases
        )))
        _children = list(itertools.chain(*(
            getattr(b, '_children', []) for b in bases
        )))

        for k, v in attributes.items():
            if isinstance(v, ConfigProperty):
                _props.append(k)
            elif isinstance(v, type) and hasattr(v, '_props'):
                # No way to check if actually a Nestable (chicken or egg?)
                # Adding a Nested class as subclass would mean to have
                # to loop over the attributes of the class and do it
                # recursively. So it needs to be a Nestable otherwise
                # the descriptor will trow because no get_root.
                _key = stringcase.snakecase(k)
                setattr(v, '_key', _key)
                _new[_key] = _NestableDescriptor(
                    f'_{_key}', getattr(v, '_props'), v, comment=v.__doc__
                )
                _children.append(v)
                _props.append(_key)

        _new['_children'] = _children
        _new['_props'] = _props

        # pylint: disable=too-many-function-args
        return abc.ABCMeta.__new__(mcs, name, bases, _new)


class Nestable(collections.abc.Mapping, metaclass=ConfigMeta):
    _parent: typing.Any
    _key: str
    _children = []
    _props = []

    def __init__(self, parent=None, parent_len=0):
        self._parent = parent
        self._parent_len = parent_len
        for child_cls in self._children:
            # noinspection PyProtectedMember
            var_name = child_cls._key
            setattr(
                self,
                var_name,
                child_cls(self, parent_len + 1)
            )

    def get_root(self, current=None):
        parent = self._parent
        if parent is None:
            return self
        levels = [current, self._key] if current else [self._key]
        last_parent = parent
        while parent is not None:
            key = getattr(parent, '_key', None)
            if key is not None:
                levels.append(key)
            last_parent = parent
            parent = getattr(parent, '_parent', None)
        return last_parent, reversed(levels)

    def __getitem__(self, k):
        # Just go into descriptor.
        return getattr(self, k)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __len__(self):  # pragma: no cover
        return len(self._props)

    def __iter__(self):
        for prop in self._props:
            yield prop

    def get_prop_paths(self, parent=''):
        children = [getattr(x, '_key') for x in self._children]
        if not parent and hasattr(self, '_key'):  # pragma: no cover
            parent = self._key
        for prop in (getattr(type(self), x) for x in self._props):
            value = getattr(self, prop.name)
            path = f'{parent + "." if parent else ""}{prop.name}'
            yield path, value, prop
            if prop.name in children:
                for k, v, p in value.get_prop_paths(path):
                    yield k, v, p


class _NestableDescriptor(ConfigProperty):

    def __init__(self, nestable, props, nested_cls, comment=None):
        default = {
            prop.name: prop.default
            for prop in (getattr(nested_cls, p) for p in props)
        }
        super().__init__(default, comment, config_type=dict)
        self.nestable = nestable
        self.name = nestable[1:]

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__.get(self.nestable, self.default)

    def __set__(self, instance, value):
        instance.__dict__[self.nestable] = value


class Config(Nestable):
    """
    Root config class, assign ConfigProperties as class members.
    """
    _serializer: BaseConfigSerializer
    _config_format: ConfigFormat

    def __init__(
            self,
            config_format: ConfigFormat = ConfigFormat.TOML,
            root_name='config'
    ):
        super().__init__(None)
        self._data = {}
        self.root_name = root_name
        self.config_format = config_format
        self._app = None

    def __getitem__(self, k):
        # Here get the prop descriptor and return the value or default
        prop = getattr(type(self), k)
        return self._data.get(k, prop.default)

    def read_dict(self, data: dict):
        self._data = data

    def read_file(self, path: str):
        self._data = self._serializer.load(path)

    def save(self, path: str):
        self._serializer.dump(self, path)

    @property
    def config_format(self) -> ConfigFormat:
        return self._config_format

    @config_format.setter
    def config_format(self, value: ConfigFormat):
        self._config_format = value
        self._serializer = self._config_format.serializer(self)


def config_factory(data, root=None, key=None):

    props = []
    children = []

    class _Current:
        pass

    for k, v in data.items():
        if isinstance(v, dict):
            nestable = config_factory(v, _Current, k)
            setattr(
                _Current, k,
                _NestableDescriptor(f'_{k}', list(v.keys()), nestable)
            )
            children.append(nestable)
        else:
            setattr(_Current, k, ConfigProperty(name=k, default=v))
        props.append(k)

    setattr(_Current, '_props', props)
    setattr(_Current, '_children', children)

    if root is None:
        class _Wrapped(_Current, Config):
            pass
    else:
        setattr(_Current, '_key', key)

        class _Wrapped(_Current, Nestable):
            pass

    return _Wrapped



import asyncio
import functools
from concurrent.futures import ThreadPoolExecutor


class AsyncExecutor:
    """Execute functions in a Pool Executor"""
    def __init__(self, loop=None, executor=None, max_workers=None):
        """
        :param loop: asyncio event loop.
        :param executor: Set to use an already existing PoolExecutor, default
            to a new ThreadPoolExecutor if not supplied.
        :param max_workers: Max workers of the created ThreadPoolExecutor.
        """
        self.loop = loop or asyncio.get_event_loop()
        self.lock = asyncio.Lock(loop=loop)
        if executor:  # pragma: no cover
            self.executor = executor
        else:
            self.executor = ThreadPoolExecutor(max_workers=max_workers)

    async def execute(self, func, *args, **kwargs):
        """
        Execute a sync function asynchronously in the executor.

        :param func: Synchronous function.
        :param args: Argument to give to the function.
        :param kwargs: Keyword arguments to give to the function
        :return:
        """
        return await self.loop.run_in_executor(
            self.executor,
            functools.partial(func, *args, **kwargs)
        )

    # pragma: no cover
    async def execute_with_lock(self, func, *args, **kwargs):
        """
        Acquire lock before executing the function.

        :param func: Synchronous function.
        :param args:
        :param kwargs:
        :return:
        """
        await self.lock.acquire()
        ret = await self.execute(func, *args, **kwargs)
        self.lock.release()
        return ret

    def wraps(self, func):
        """
        Wraps a synchronous function to execute in the pool when called,
        making it async.

        :param func: The function to wraps
        :return: Async wrapped function.
        """

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            return await self.execute(func, *args, **kwargs)

        return wrapper



import asyncio
import sys
import math

from colorama import Style, Fore


def colorize(text, bg=None, fg=None, style=None):
    fg = fg or ''
    bg = bg or ''
    style = style or ''
    return f'{bg}{fg}{style}{text}{Style.RESET_ALL}'


_spinner_symbols = (
    '|', '/', '-', '\\', '|', '/', '-', '\\'
)


async def spinner(condition,
                  sleep_time=0.25, message='',
                  fg=Fore.WHITE, bg=None,
                  symbols=_spinner_symbols):
    i = 0

    num_symbols = len(symbols)

    while not condition():
        symbol = symbols[i % num_symbols]
        i += 1
        if callable(message):
            msg = message()
        else:  # pragma: no cover
            msg = message
        m = colorize(f'{msg} {symbol}', fg=fg, bg=bg)
        msg = f'\r\x1b[K{m}'
        print(msg, end='', flush=True, file=sys.stderr)
        await asyncio.sleep(sleep_time)


def chunk_list(data, n):
    for i in range(0, len(data), n):
        yield data[i:i+n]


def format_table(data, formatting=None):
    # 2 padding + '|'
    max_len = max(len(x) for x in data) + 3
    if max_len % 2 == 0:  # pragma: no cover
        # Revert uneven for a bit nicer look.
        max_len += 1
    min_len = len(data) * max_len
    r = 79 / max_len

    if r - int(r) > 0:
        col = int(r) - 1
    else:  # pragma: no cover
        col = int(r)

    if min_len < 79:  # pragma: no cover
        row_len = min_len + len(data) + 1
    else:
        row_len = max_len * col
        # Add the columns + 1 for the first '|'
        row_len += col + 1

    if not formatting:
        formatting = lambda e: e  # noqa: E731

    chunks = chunk_list(data, col)
    rows = ['-' * row_len]
    for chunk in chunks:
        rows.append(
            '|{}|'.format(
                '|'.join(
                    formatting(c.center(max_len))
                    for c in chunk
                )
            )
        )
        rows.append('-' * row_len)
    return rows


def print_table(data, formatting=None, file=sys.stdout):
    print('\n'.join(format_table(data, formatting,)), file=file)


def goto_xy(stream, x, y):  # pragma: no cover
    # Make sure colorama is init on windows.
    print('%c[%d;%df' % (0x1B, y, x), end='', file=stream)


async def progress_bar(
        value_func, max_value,
        value_formatter=None,
        include_value=True,
        dents=50,
        sleep_time=0.005,
        loop=None,
        file=None,
        full_symbol='#',
        empty_symbol='-',
        start_symbol='[',
        end_symbol=']',
):
    loop = loop or asyncio.get_event_loop()
    value = 0

    value_formatter = value_formatter or (lambda x, y: f'{x} / {y}')

    while value < max_value:
        value = value_func()
        await asyncio.sleep(sleep_time, loop=loop)

        progress = math.ceil(value / max_value * dents)
        progress_line = (
            (progress * full_symbol) + (dents - progress) * empty_symbol
        )
        out = f'\r\x1b[K{start_symbol}{progress_line}{end_symbol}'
        if include_value:
            out += value_formatter(value, max_value)

        print(out, end='', flush=True, file=file)



import enum
import time


from .._tools import AutoNameEnum
from .._immutable import ImmutableDict


class Event:
    """
    Event with payload and stop property.
    """
    def __init__(self, name: str, payload: dict):
        self.name = name
        self.stop: bool = False
        self.num = 0
        self.payload = ImmutableDict(**payload)
        self.timestamp = time.time()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name} {self.timestamp}>'


class PreceptEvent(AutoNameEnum):
    """Precept cli events."""
    BEFORE_CLI_START = enum.auto()
    CLI_PARSED = enum.auto()
    CLI_STARTED = enum.auto()
    CLI_STOPPED = enum.auto()

    # pylint: disable=comparison-with-callable
    def __eq__(self, other):
        if isinstance(other, PreceptEvent):
            return self.value == other.value
        if isinstance(other, str):
            return other in (self.value, self.name, str(self))
        if isinstance(other, Event):
            return other.name == self.value
        return False

    def __hash__(self):
        return hash(self.value)

    def __str__(self):
        return self.value



class Plugin:
    """
    Plugin's are automatically added to a precept
    application upon installation.

    Set the entry point in setup to register the plugin:

        entry_point = {'{app_name}.plugins': ['plugin = my_plugin:plugin']}
    """
    name: str = ''

    async def setup(self, application):
        """
        Setup the plugin

        :param application: The running precept application.
        :type application: precept.Precept
        :return:
        """
        raise NotImplementedError



import sys
from enum import Enum

__all__ = [
    'is_windows',
    'AutoNameEnum'
]


def is_windows():
    return sys.platform == 'win32'


class AutoNameEnum(Enum):
    # noinspection PyMethodParameters
    # pylint: disable=no-self-argument, unused-argument, no-member
    def _generate_next_value_(name, *args):
        return name.lower()



"""Class based cli builder with sub commands."""
import argparse
import asyncio
import itertools
import typing
import functools
import inspect

import stringcase

from ._services import Service
from ._immutable import ImmutableDict
from .events import PreceptEvent


def _flags_key(flags):
    return stringcase.snakecase(flags[-1].lstrip('-'))


class Argument(ImmutableDict):
    """
    Argument of a Command, can either be optional or not depending on the flags

    .. seealso::

        https://docs.python.org/3/library/argparse.html#the-add-argument-method

    """

    # pylint: disable=unused-argument, redefined-builtin
    def __init__(
            self,
            *flags: str,
            type: type = None,
            help: str = None,
            choices: typing.Iterable = None,
            default: typing.Any = None,
            nargs: typing.Union[str, int] = None,
            action: str = None,
            required: bool = None,
            metavar: str = None,
            dest: str = None,
    ):
        """
        :param flags: How to call the argument, prefixing with ``-`` makes the
            argument a keyword. Example: ``'-d', '--date'`` make the variable
            available as ``date``, but can be supplied as ``-d`` from the cli.
        :param type: The type of the variable to cast to, default to str.
        :param help: Description to go along with
        :param choices: The available choices to choose from.
        :param default: The value to take if not supplied.
        :param nargs: Number of times the argument can be supplied.
        :param action: What to do with the argument.
        :param required: Makes a keyword argument required.
        :param metavar: Name in help.
        :param dest: The name of the variable to add the value to once parsed.
        """
        super().__init__(**{
            k: v for k, v in locals().items()
            if k in self._prop_keys and v is not None
        })

    def register(self, parser):
        options = {k: v for k, v in self.items() if k != 'flags'}

        if 'default' in options and 'help' not in options:
            # Otherwise the default value don't show up.
            options['help'] = '-'

        flags = self.flags
        if len(self.flags) == 1 and not self.flags[0].startswith('-'):
            flags = [stringcase.snakecase(flags[0])]

        parser.add_argument(*flags, **options)

    @property
    def flag_key(self):
        return _flags_key(self.flags)


class CombinedFormatter(argparse.ArgumentDefaultsHelpFormatter,
                        argparse.RawDescriptionHelpFormatter):
    pass


class Command:
    """
    Command decorator, methods of `CliApp` subclasses decorated with this gets
    a sub-command in the parser.

    Wrapped methods will gets the arguments by the ``Argument`` flag.
    """
    arguments: typing.Iterable[Argument]
    description: str

    # pylint: disable=redefined-builtin
    def __init__(
            self,
            *arguments: Argument,
            name: str = None,
            description: str = None,
            help: str = None,
            auto: bool = False,
            services: typing.List[Service] = None
    ):
        self.arguments = arguments
        self._command_name = None
        self.command_name = name
        self.description = description
        self.help = help
        self._wrapped = None
        self.obj_name = None
        self.auto = auto
        self.services = services

    def __call__(self, obj):
        self.obj_name = getattr(obj, '__name__')
        if not self.command_name:
            self.command_name = self.obj_name

        if not self.description:
            self.description = getattr(obj, '__doc__', '')

        if isinstance(obj, type):

            # pylint: disable=used-before-assignment
            class NewCommand(obj, CommandClass, metaclass=CommandMeta):
                command = self

            self._wrapped = NewCommand()
        else:
            if self.auto:
                arguments = []
                signature = inspect.signature(obj)
                for k, v in signature.parameters.items():
                    if k in ('self', 'args', 'kwargs'):
                        continue
                    key = stringcase.spinalcase(k)
                    default = None
                    _type = None
                    if v.annotation:
                        _type = v.annotation
                    if v.default is not v.empty:
                        key = f'--{key}'
                        default = v.default
                        if _type is None:
                            _type = type(default)
                    arguments.append(
                        Argument(key, type=_type, default=default)
                    )
                self.arguments = tuple(arguments) + tuple(self.arguments)

            self._wrapped = CommandFunction(obj, self)

        return self._wrapped

    def register(self, subparsers):
        parser = subparsers.add_parser(
            self.command_name,
            description=self.description,
            help=self.help or self.description,
            formatter_class=CombinedFormatter
        )
        if isinstance(self._wrapped, CommandClass):
            subs = parser.add_subparsers(
                title='Commands', dest='command'
            )
            # noinspection PyProtectedMember
            for command_name, command, _ in self._wrapped.get_commands():
                # get_commands return itself so make sure it's not the same.
                if command_name != self.command_name:
                    command.register(subs)
        for arg in self.arguments:
            arg.register(parser)

    @property
    def command_name(self):
        return self._command_name

    @command_name.setter
    def command_name(self, value):
        self._command_name = stringcase.spinalcase(value) if value else value

    def __hash__(self):
        return self.command_name


class WrappedCommand:
    command: Command

    def __repr__(self):  # pragma: no cover
        return f'<{self.__class__.__name__} "{self.command.command_name}">'

    def get_commands(self) -> typing.Tuple[str, Command, typing.Callable]:
        raise NotImplementedError


class CommandFunction(WrappedCommand):
    def __init__(self, func, command):
        self.command = command
        self.func = functools.wraps(self)(func)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __get__(self, instance, owner):
        # Allow to use the self of the class
        if instance is None:
            return self
        return functools.partial(self.__call__, instance)

    def get_commands(self):
        return [(self.command.command_name, self.command, self)]

    def set_func(self, func):
        self.func = func


class CommandClass(WrappedCommand):
    _commands: typing.List[str]

    def __call__(self, command, *args, **kwargs):
        return getattr(self, command)(*args, **kwargs)

    def get_commands(self):
        c = []
        for command in (getattr(self.__class__, x) for x in self._commands):
            if issubclass(command.__class__, CommandClass):
                # Recursively get all the commands.
                c += command.get_commands()
            else:
                # Get the true one.
                c.append((
                    command.command.command_name,
                    command.command,
                    self.clean_arguments(
                        getattr(self, command.command.obj_name))
                ))

        # noinspection PyTypeChecker
        return c + [(self.command.command_name, self.command, self)]

    def clean_arguments(self, func):

        @functools.wraps(func)
        def wrap_arguments(*args, **kwargs):
            to_clean = []

            for argument in self.command.arguments:
                key = argument.flag_key
                value = kwargs.get(key)
                setattr(self, key, value)
                to_clean.append(key)

            cleaned = {k: v for k, v in kwargs.items() if k not in to_clean}
            return func(*args, **cleaned)

        return wrap_arguments


class CommandMeta(type):
    def __new__(mcs, name, bases, attributes):
        new_attributes = dict(**attributes)
        commands = []

        for k, v in itertools.chain(*(
                z.items() for z in [y.__dict__ for y in bases] + [attributes]
        )):
            if isinstance(v, WrappedCommand):
                commands.append(k)
            if isinstance(v, CommandClass):
                # The ast get to nested's first so this can work.
                commands += v._commands

        new_attributes['_commands'] = commands

        return type.__new__(mcs, name, bases, new_attributes)


class Cli:
    """argparse cli wrapper."""

    def __init__(self, *commands,
                 prog='',
                 description='',
                 formatter_class=CombinedFormatter,
                 global_arguments=None,
                 default_command=None,
                 events=None,
                 on_parse=None):
        self.prog = prog
        self.default_command = default_command
        self.parser = argparse.ArgumentParser(
            prog=self.prog,
            description=description,
            formatter_class=formatter_class,
        )

        self._global_arguments = global_arguments or []
        self.globals = {}
        self._on_parse = on_parse
        self._events = events

        for g in self._global_arguments:
            g.register(self.parser)

        self.commands = {}

        subparsers = self.parser.add_subparsers(
            title='Commands', dest='command', metavar=''
        )

        for command_name, command, wrapper in commands:
            command.register(subparsers)
            self.commands[command_name] = wrapper

    async def run(self, args=None):
        """
        Parse and call the appropriate handler.

        :param args: None takes sys.argv
        :return:
        """
        namespace = self.parser.parse_args(args=args)
        command = self.commands.get(namespace.command)
        kw = vars(namespace).copy()
        kw.pop('command')

        for g in self._global_arguments:
            key = _flags_key(g.flags)
            self.globals[key] = kw.pop(key)

        if callable(self._on_parse):
            await self._on_parse(namespace)

        if command:
            operation = command(**kw)
            event = self._events.dispatch(
                str(PreceptEvent.CLI_STARTED), command=namespace.command
            )
            await asyncio.gather(event, operation)
        elif self.default_command:
            await self.default_command(**vars(namespace))
        else:  # pragma: no cover
            self.parser.print_help()



import logging
import sys

from colorama import Fore, Style

from .console import colorize

_colors = {
    'INFO': {
        'fg': Fore.LIGHTBLUE_EX,
        'style': Style.BRIGHT,
    },
    'DEBUG': {
        'fg': Fore.LIGHTBLACK_EX,
        'style': Style.BRIGHT,
    },
    'ERROR': {
        'fg': Fore.RED,
        'style': Style.BRIGHT,
    },
    'WARNING': {
        'fg': Fore.YELLOW,
        'style': Style.BRIGHT,
    },
}


class ColorFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, colors=None, style='%'):
        super(ColorFormatter, self).__init__(
            fmt=fmt, datefmt=datefmt, style=style
        )
        self.colors = colors or _colors

    def format(self, record: logging.LogRecord):
        formatted = super(ColorFormatter, self).format(record)
        return colorize(
            f'\r\x1b[K{formatted}',
            **self.colors.get(record.levelname)
        )


def setup_logger(
        logger_name,
        level=logging.INFO,
        fmt=None,
        datefmt=None,
        stream=sys.stderr,
        colors=None,
        style='%',
):
    logger = logging.getLogger(logger_name)
    if not logger.handlers:
        std_handler = logging.StreamHandler(stream=stream)
        std_handler.setFormatter(ColorFormatter(fmt, datefmt, colors, style))
        logger.addHandler(std_handler)
    logger.setLevel(level)
    logger.propagate = False

    return logger



from ._version import __version__  # noqa: F401
from ._cli import Command, Argument  # noqa: F401
from ._precept import Precept  # noqa: F401
from ._tools import AutoNameEnum, is_windows  # noqa: F401, F403
from ._immutable import ImmutableProp, ImmutableDict, ImmutableMeta  # noqa: F401, F403, E501
from ._configs import (  # noqa: F401
    ConfigProperty, Config, ConfigFormat, Nestable, config_factory
)
from ._executor import AsyncExecutor  # noqa: F401
from ._services import Service  # noqa: F401
from ._plugins import Plugin  # noqa: F401


__all__ = [
    '__version__',
    'Command',
    'Argument',
    'Precept',
    'ImmutableDict',
    'ImmutableMeta',
    'ImmutableProp',
    'ConfigProperty',
    'Config',
    'ConfigFormat',
    'Nestable',
    'config_factory',
    'AsyncExecutor',
    'Service',
    'Plugin',
    'is_windows',
    'AutoNameEnum'
]



import functools
import stringcase

from .events import EventDispatcher


def _dispatch_wrap(func, event, running):

    @functools.wraps(func)
    async def dispatcher(self, *args, **kwargs):
        self.running = running
        payload = await func(self, *args, **kwargs)
        await self.events.dispatch(event, instance=self)
        return payload

    return dispatcher


class ServiceMeta(type):
    def __new__(mcs, name, bases, attributes):
        _new = dict(**attributes)
        _name = attributes.get('name') or stringcase.snakecase(name)
        _new['name'] = _name

        for _method, _running in (
                ('setup', False), ('start', True), ('stop', False)
        ):
            _func = attributes.get(_method)
            if _func:
                _new[_method] = _dispatch_wrap(
                    _func, f'{_name}_{_method}', _running
                )

        return type.__new__(mcs, name, bases, _new)


class Service(metaclass=ServiceMeta):
    """
    Service's runs alongside the main application.

    Communicate via events.

    :Events:
        - ``{name}_setup`` when added to services.
        - ``{name}_start`` after calling start.
        - ``{name}_stop`` after calling stop.

    """
    name: str = None

    def __init__(self, events: EventDispatcher = None):
        self.events = events or EventDispatcher()
        self.running: bool = False
        self.ready: bool = False

    async def setup(self, application):
        """
        Called when added to services.

        Use this to set events and other post initialization that may need
        the application instance.

        :param application: Precept application
        :type application: precept.Precept
        :return:
        """
        raise NotImplementedError

    async def start(self):
        """
        Start the service.

        :return:
        """
        raise NotImplementedError

    async def stop(self):
        """
        Stop the service.

        :return:
        """
        raise NotImplementedError



import asyncio
import string
import sys
import threading
from itertools import chain

from precept._tools import is_windows


class Key:  # pragma: no cover
    def __init__(self, value, clean=None):
        self.value = value
        self.clean = clean

    def __str__(self):
        return self.clean or self.value

    def __eq__(self, other):
        if isinstance(other, Key):
            return other.value == self.value
        if isinstance(other, str):
            return other == self.value
        return False

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return f"<Key '{self.clean or self.value}'>"


class Keys:  # pragma: no cover
    SPACE = Key(' ', 'space')
    BACKSPACE = Key('\x7f', 'backspace')
    ENTER = Key('\r', 'enter')
    ESCAPE = Key('\x1b', 'escape')
    INSERT = Key('\x1b[2~', 'insert')
    END = Key('\x1b[F', 'end')
    HOME = Key('\x1b[H', 'home')
    DELETE = Key('\x1b[3~', 'delete')
    DOWN = Key('\x1b[B', 'down')
    UP = Key('\x1b[A', 'up')
    LEFT = Key('\x1b[D', 'left')
    RIGHT = Key('\x1b[C', 'right')

    F1 = Key('\x1bOP', 'F1')
    F2 = Key('\x1bOQ', 'F2')
    F3 = Key('\x1bOR', 'F3')
    F4 = Key('\x1bOS', 'F4')
    F5 = Key('\x1bO15~', 'F5')
    F6 = Key('\x1bO17~', 'F6')
    F7 = Key('\x1bO18~', 'F7')
    F8 = Key('\x1bO19~', 'F8')
    F9 = Key('\x1bO20~', 'F9')
    F10 = Key('\x1bO21~', 'F10')
    F11 = Key('\x1bO23~', 'F11')
    F12 = Key('\x1bO24~', 'F12')

    CTRL_C = Key('\x03', 'ctrl-c')
    CTRL_A = Key('\x01', 'ctrl-a')
    CTRL_ALT_A = Key('\x1b\x01', 'ctrl-alt-a')
    CTRL_ALT_DEL = Key('\x1b[3^', 'ctrl-alt-del')
    CTRL_B = Key('\x02', 'ctrl-b')
    CTRL_D = Key('\x04', 'ctrl-d')
    CTRL_E = Key('\x05', 'ctrl-e')
    CTRL_F = Key('\x06', 'ctrl-f')
    CTRL_Z = Key('\x1a', 'ctrl-z')

    SPECIAL_KEYS = (
        SPACE, BACKSPACE, ENTER, ESCAPE, INSERT, END, HOME,
        DELETE, DOWN, UP, LEFT, RIGHT,
        F1, F2, F3, F4, F5, F6, F7, F8, F9, F10, F11, F12,
        CTRL_C, CTRL_A, CTRL_ALT_A, CTRL_ALT_DEL, CTRL_B,
        CTRL_D, CTRL_E, CTRL_F, CTRL_Z
    )
    keys = {
        x: Key(x) for x in chain(string.ascii_letters, string.digits)
    }
    keys.update({
        x.value: x for x in SPECIAL_KEYS
    })

    @classmethod
    def get_key(cls, value, default=None):
        return cls.keys.get(value) or default


class GetChar:
    def __init__(self):  # pragma: no cover
        if is_windows():
            # pylint: disable=import-error
            import msvcrt
            self.get_char = msvcrt.getwch
        else:
            import termios
            import tty

            def get_char():
                # pylint: disable=assignment-from-no-return
                fileno = sys.stdin.fileno()
                old = termios.tcgetattr(fileno)

                try:
                    tty.setraw(fileno)
                    raw = termios.tcgetattr(fileno)
                    raw[1] = old[1]
                    termios.tcsetattr(fileno, termios.TCSADRAIN, raw)
                    char = sys.stdin.read(1)
                finally:
                    termios.tcsetattr(fileno, termios.TCSADRAIN, old)
                return char

            self.get_char = get_char

    def __call__(self):  # pragma: no cover
        return self.get_char()

    async def deferred(self, executor):  # pragma: no cover
        return await executor.execute(self.get_char)


getch = GetChar()


class KeyHandler:  # pragma: no cover
    def __init__(self, handlers, loop=None, default_handler=None):  # pragma: no cover # noqa: E501
        self.handlers = handlers
        self.handlers.update({
            Keys.CTRL_C: lambda _, stop: stop(),
        })
        self.default_handler = default_handler
        self.loop = loop or asyncio.get_event_loop()
        self.queue = asyncio.Queue(loop=self.loop)
        self.stop_event = asyncio.Event(loop=self.loop)
        self._consumer = None
        self._producer = None

    def stop(self):  # pragma: no cover
        self.stop_event.set()

    def read(self):  # pragma: no cover
        # Make non-blocking.
        while not self.stop_event.is_set():
            char = getch()
            asyncio.ensure_future(self.queue.put(char), loop=self.loop)

    async def handle(self):  # pragma: no cover
        while not self.stop_event.is_set():
            await asyncio.sleep(0.00001)
            try:
                msg = self.queue.get_nowait()
            except asyncio.QueueEmpty:
                pass
            else:
                handler = self.handlers.get(msg)
                if handler:
                    handler(msg, self.stop)
                elif self.default_handler:
                    self.default_handler(msg, self.stop)

    async def __aenter__(self):  # pragma: no cover
        self._producer = threading.Thread(target=self.read)
        self._producer.daemon = True
        self._producer.start()
        self._consumer = asyncio.ensure_future(self.handle(), loop=self.loop)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        self.stop()
        await self._consumer

    def print_keys(self, file=sys.stdout):  # pragma: no cover
        for k, v in self.handlers.items():
            doc = getattr(v, '__doc__', getattr(v, '__name__', ''))
            clean_key = Keys.get_key(k, k)
            if k == Keys.CTRL_C:
                continue
            print(f'{clean_key}: {doc}', file=file)


if __name__ == '__main__':  # pragma: no cover
    main_loop = asyncio.get_event_loop()

    async def main():

        namespace = {
            'i': 0
        }

        def hand(msg, stop):
            namespace['i'] += 1
            if namespace['i'] >= 10:
                stop()
            print(repr(msg), file=sys.stderr)

        async with KeyHandler({}, default_handler=hand, loop=main_loop) as k:
            k.print_keys()
            print('Type 10 chars')
            while not k.stop_event.is_set():
                print('.', end='', flush=True)
                await asyncio.sleep(1, loop=main_loop)

    main_loop.run_until_complete(main())



class PreceptError(Exception):
    """Base exception thrown by precept"""


class ConfigError(PreceptError):
    """Error in the config system."""


class ImmutableError(PreceptError):
    """Immutable properties cannot change"""



import collections

from ._event import Event


class EventDispatcher:
    """Dispatch events to subscribers functions."""
    def __init__(self):
        self._subscribers = collections.defaultdict(list)

    def subscribe(self, event: str, func):
        """
        Subscribe func to execute every time event is dispatched.

        :param event: The event to subscribe to.
        :param func: The func to call when event is dispathed.
        :return:
        """
        self._subscribers[event].append(func)

    async def dispatch(self, event: str, **payload):
        """
        Dispatch an event with optional payload data.

        :param event: Name of the event.
        :param payload: Data of the event.
        :return:
        """
        action = Event(event, payload=payload)
        for subscriber in self._subscribers[event]:
            await subscriber(action)
            if action.stop:
                break
            action.num += 1
